# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp
from jax import random, jit
from cr.nimble import is_matrix


def randomize_rows(key, X):
    """Randomizes the rows in X

    Args:
        key: a PRNG key used as the random key.
        X (jax.numpy.ndarray): A 2D data matrix

    Returns:
        (jax.numpy.ndarray): The data matrix with randomized rows
    """
    assert is_matrix(X)
    m, n = X.shape
    r = random.permutation(key, m)
    return X[r, :]

def randomize_cols(key, X):
    """Randomizes the columns in X

    Args:
        key: a PRNG key used as the random key.
        X (jax.numpy.ndarray): A 2D data matrix

    Returns:
        (jax.numpy.ndarray): The data matrix with randomized columns
    """
    assert is_matrix(X)
    m, n = X.shape
    r = random.permutation(key, n)
    return X[:, r]



def largest_indices(x, K):
    """Returns the indices of K largest entries in x by magnitude

    Args:
        x (jax.numpy.ndarray): An data vector/point
        K (int): The number of largest entries to be identified in x

    Returns:
        (jax.numpy.ndarray): An index vector of size K identifying the K largest entries in x
        in descending order
    """
    indices = jnp.argsort(jnp.abs(x))
    return indices[:-K-1:-1]

def largest_indices_rw(X, K):
    """Returns the indices of K largest entries by magnitude in each row of X

    Args:
        X (jax.numpy.ndarray): An (S,N) data matrix with data points in rows
        K (int): The number of largest entries to be identified in each row of X

    Returns:
        (jax.numpy.ndarray): An (S,K) index matrix indices of K largest elements in each row of X
    """
    indices = jnp.argsort(jnp.abs(X), axis=1)
    return indices[:, :-K-1:-1]

def largest_indices_cw(X, K):
    """Returns the indices of K largest entries by magnitude in each column of X

    Args:
        X (jax.numpy.ndarray): An (N,S) data matrix with data points in columns
        K (int): The number of largest entries to be identified in each column of X

    Returns:
        (jax.numpy.ndarray): An (K,S) index matrix indices of K largest elements in each column of X
    """
    indices = jnp.argsort(jnp.abs(X), axis=0)
    return indices[:-K-1:-1, :]

def take_along_rows(X, indices):
    """Picks K entries from each row of X specified by indices matrix

    Args:
        X (jax.numpy.ndarray): An (S,N) data matrix with data points in rows
        indices (jax.numpy.ndarray): An (S,K) index matrix identifying the values to be picked up from X

    Returns:
        (jax.numpy.ndarray): An (S,K) data matrix subset of X containing K elements from each row of X
    """
    return jnp.take_along_axis(X, indices, axis=1)

def take_along_cols(X, indices):
    """Picks K entries from each column of X specified by indices matrix

    Args:
        X (jax.numpy.ndarray): An (N,S) data matrix with data points in columns
        indices (jax.numpy.ndarray): An (K,S) index matrix identifying the values to be picked up from X

    Returns:
        (jax.numpy.ndarray): An (K,S) data matrix subset of X containing K elements from each column of X
    """
    return jnp.take_along_axis(X, indices, axis=0)


def sparse_approximation(x, K):
    """Keeps only largest K non-zero entries by magnitude in a vector x

    Args:
        x (jax.numpy.ndarray): An data vector/point
        K (int): The number of largest entries to be kept in x

    Returns:
        (jax.numpy.ndarray): x modified so that all entries except the K largest entries are set to 0
    """
    if K == 0:
        return x.at[:].set(0)
    indices = jnp.argsort(jnp.abs(x))
    #print(x, K, indices)
    return x.at[indices[:-K]].set(0)
    
def sparse_approximation_cw(X, K):
    #return jax.vmap(sparse_approximation, in_axes=(1, None), out_axes=1)(X, K)
    """Keeps only largest K non-zero entries by magnitude in each column of X

    Args:
        X (jax.numpy.ndarray): An (N,S) data matrix with data points in columns
        K (int): The number of largest entries to be kept in each column of X

    Returns:
        (jax.numpy.ndarray): X modified so that all entries except the K largest entries are set to 0 in each column
    """
    if K == 0:
        return X.at[:].set(0)
    indices = jnp.argsort(jnp.abs(X), axis=0)
    for c in range(X.shape[1]):
        ind = indices[:-K, c]
        X = X.at[ind, c].set(0)
    return X

def sparse_approximation_rw(X, K):
    """Keeps only largest K non-zero entries by magnitude in each row of X

    Args:
        X (jax.numpy.ndarray): An (S,N) data matrix with data points in rows
        K (int): The number of largest entries to be kept in each row of X

    Returns:
        (jax.numpy.ndarray): X modified so that all entries except the K largest entries are set to 0 in each row
    """
    if K == 0:
        return X.at[:].set(0)
    indices = jnp.argsort(jnp.abs(X), axis=1)
    for r in range(X.shape[0]):
        ind = indices[r, :-K]
        X = X.at[r, ind].set(0)
    return X

def build_signal_from_indices_and_values(length, indices, values):
    """Builds a sparse signal from its non-zero entries (specified by their indices and values)

    Args:
        length (int): Length of the sparse signal
        indices (jax.numpy.ndarray): An index vector of length K identifying non-zero entries
        values (jax.numpy.ndarray): Values to be stored in the non-zero positions

    Returns:
        (jax.numpy.ndarray): Resulting sparse signal such that x[indices] == values 
    """
    x = jnp.zeros(length)
    indices = jnp.asarray(indices)
    values = jnp.asarray(values)
    return x.at[indices].set(values)


def nonzero_values(x):
    """Returns the values of non-zero entries in x

    Args:
        x (jax.numpy.ndarray): A sparse signal

    Returns:
        (jax.numpy.ndarray): The signal stripped of its zero values


    Note:
        This function cannot be JIT compiled as the size of output is data dependent.
    """
    return x[x != 0]

def nonzero_indices(x):
    """Returns the indices of non-zero entries in x

    Args:
        x (jax.numpy.ndarray): A sparse signal

    Returns:
        (jax.numpy.ndarray): The indices of nonzero entries in x

    Note:
        This function cannot be JIT compiled as the size of output is data dependent.

    See Also:
        :func:`support`
    """
    return jnp.nonzero(x)[0]


def support(x):
    """Returns the indices of non-zero entries in x

    Args:
        x (jax.numpy.ndarray): A sparse signal

    Returns:
        (jax.numpy.ndarray): The support of x a.k.a. the indices of nonzero entries in x
        
    Note:
        This function cannot be JIT compiled as the size of output is data dependent.

    See Also:
        :func:`nonzero_indices`
    """
    return jnp.nonzero(x)[0]






# Copyright 2022 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Run Length Encoding of Binary Maps


References:

https://stackoverflow.com/questions/1066758/find-length-of-sequences-of-identical-values-in-a-numpy-array-run-length-encodi
"""

from bitarray import bitarray
from bitarray.util import int2ba, ba2int
import numpy as np

def count_binary_runs(input_arr):
    """Returns the runs of 0s and 1s in a binary map
    """
    # make sure that it is a numpy array
    input_arr = np.asarray(input_arr)
    # the first bit
    b = input_arr[0]
    # the last bit
    e = input_arr[-1]
    # extend the array
    extended = np.hstack(([1-b], input_arr, [1 - e]))
    # locate the changes
    diffs = np.diff(extended)
    markers, = np.where(diffs)
    runs = np.diff(markers)
    return runs

B00 = bitarray('00')
B01 = bitarray('01')
B10 = bitarray('10')
B11 = bitarray('11')
NUM_BITS_RUN_LEN = 4


def encode_binary_arr(input_arr):
    """Encodes a binary array into a bit array via run length encoding
    """
    # the first bit
    b = input_arr[0]
    # the runs
    runs = count_binary_runs(input_arr)
    # build the bit array
    a = bitarray()
    a.append(b)
    for run in runs:
        run = int(run)
        if run == 1:
            a.extend(B00)
            continue
        if run == 2:
            a.extend(B01)
            continue
        if run == 3:
            a.extend(B10)
            continue
        # run is 4 or more
        a.extend(B11)
        # now record number of bits for the run
        bl = run.bit_length()
        a.extend(int2ba(bl, NUM_BITS_RUN_LEN))
        # now record the run
        a.extend(int2ba(run))
    return a



def decode_binary_arr(input_bit_arr : bitarray):
    """Decodes a binary array from a bit array via run length decoding
    """
    a = input_bit_arr
    result = []
    # The first bit
    b = a[0]
    idx = 1
    # number of bits in the encoded bit array
    n = len(a)
    while idx < n:
        # read the next 2 bits
        code = a[idx:idx+2]
        idx += 2
        code = ba2int(code)
        run = code + 1
        if code == 3:
            # we need to decode run from the stream
            bl = ba2int(a[idx:idx+NUM_BITS_RUN_LEN])
            idx += NUM_BITS_RUN_LEN
            run = ba2int(a[idx:idx+bl])
            idx += bl
        for i in range(run):
            result.append(b)
        b = 1 - b
    return np.array(result)


def binary_compression_ratio(input_arr, output_arr, bits_per_sample=1):
    """Returns the compression ratio of binary array compression algorithm
    """
    out_len = output_arr.nbytes * 8
    ratio = len(input_arr) * bits_per_sample / out_len
    return ratio

def binary_space_saving_ratio(input_arr, output_arr, bits_per_sample=1):
    """Returns the space saving ratio of binary array compression algorithm
    """
    out_len = output_arr.nbytes * 8
    return 1 - out_len / (len(input_arr) * bits_per_sample)



# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Signal Processing Utilities
"""

# pylint: disable=W0611


from cr.nimble._src.dsp.convolution import (
    # convolution
    vec_convolve,
    vec_convolve_jit,
)

# Energy
from cr.nimble._src.dsp.energy import (

    # energy of a signal
    energy,
    find_first_signal_with_energy_le_rw,
    find_first_signal_with_energy_le_cw,
)

# Thresholding
from cr.nimble._src.dsp.thresholding import (

    hard_threshold,
    hard_threshold_sorted,
    hard_threshold_by,
    largest_indices_by,
    energy_threshold,
)

# Scaling
from cr.nimble._src.dsp.scaling import (

    scale_to_0_1,
    descale_from_0_1,
    # statistical normalization of data
    scale_0_mean_1_var,
    scale_0_mean_1_var_jit,
)

# Quantization
from cr.nimble._src.dsp.quantization import (
    quantize_1,
    inv_quantize_1,
)


# Spectrum
from cr.nimble._src.dsp.spectrum import (
    norm_freq,
    frequency_spectrum,
    power_spectrum
)

# Interpolation
from cr.nimble._src.dsp.interpolation import (
    # interpolate via fourier transform
    interpft,
)

# Signal Features
from cr.nimble._src.dsp.features import (
    dynamic_range,
    nonzero_dynamic_range,
)

# Sparse Signals
from cr.nimble._src.dsp.sparse import (
    nonzero_values,
    nonzero_indices,
    support,
    largest_indices,
    sparse_approximation,
    build_signal_from_indices_and_values,
)


# Sparse Signal Matrices
from cr.nimble._src.dsp.sparse import (
    randomize_rows,
    randomize_cols,
    # row wise
    take_along_rows,
    largest_indices_rw,
    sparse_approximation_rw,
    # column wise
    take_along_cols,
    largest_indices_cw,
    sparse_approximation_cw,
)

# Signal Comparison
from cr.nimble._src.signalcomparison import (
    SignalsComparison,
    snrs_cw,
    snrs_rw,
    snr
)

# Noise
from cr.nimble._src.noise import (
    awgn_at_snr
)

# Discrete Cosine Transform
from cr.nimble._src.dsp.dct import (
    dct,
    idct,
    orthonormal_dct,
    orthonormal_idct
)


# Walsh Hadamard
from cr.nimble._src.dsp.wht import (
    fwht,
)

from cr.nimble._src.dsp.synthetic_signals import (
    time_values,
)



# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
from jax import jit
import jax.numpy.fft as jfft

def interpft(x, N):
    """Interpolates x to n points in Fourier Transform domain
    """
    n = len(x)
    assert n < N
    a = jfft.fft(x)
    nyqst = (n + 1) // 2
    z = jnp.zeros(N -n)
    a1 = a[:nyqst+1]
    a2 = a[nyqst+1:]
    b = jnp.concatenate((a1, z, a2))
    if n % 2 == 0:
        b = b.at[nyqst].set(b[nyqst] /2 )
        b = b.at[nyqst + N -n].set(b[nyqst])
    y = jfft.ifft(b)
    if jnp.isrealobj(x):
        y = jnp.real(y)
    # scale it up
    y = y * (N / n)
    return y




# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
from jax import jit
import jax.numpy.fft as jfft

from cr.nimble import next_pow_of_2

def norm_freq(frequency, sampling_rate):
    """Returns the normalized frequency

    The Nyquist frequency is the half of the sampling rate.
    In normalized range, the Nyquist frequency has a value of 1.
    If sampling rate is 200 Hz and signal frequency is
    20 Hz, then Nyquist frequency is 100 Hz and the
    normalized frequency is 0.2.

    Args:
        frequency (float): Frequency in Hz.
        sampling_rate (float): Sampling rate of signal in Hz.

    Returns:
        float: Normalized sampling frequency
    """
    return 2.0 * frequency / sampling_rate


def frequency_spectrum(x, dt=1.):
    """Frequency spectrum of 1D data using FFT
    """
    n = len(x)
    nn = next_pow_of_2(n)
    X = jfft.fft(x, nn)
    f = jfft.fftfreq(nn, d=dt)
    X = jfft.fftshift(X)
    f = jfft.fftshift(f)
    return f, X

def power_spectrum(x, dt=1.):
    """Power spectrum of 1D data using FFT
    """
    n = len(x)
    T = dt * n
    f, X = frequency_spectrum(x, dt)
    nn = len(f)
    n2 = nn // 2
    f = f[n2:]
    X = X[n2:]
    sxx = (X * jnp.conj(X)) / T
    sxx = jnp.abs(sxx)
    return f, sxx



# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
from jax import jit


def dynamic_range(x):
    """Returns the ratio of largest and smallest values (by magnitude) in x (dB)

    Args:
        x (jax.numpy.ndarray): A signal

    Returns:
        (float): The dynamic range between largest and smallest value

    Note:
        This function is not suitable for sparse signals where some values are actually 0

    See Also:
        :func:`nonzero_dynamic_range`
    """
    x = jnp.sort(jnp.abs(x))
    return 20 * jnp.log10(x[-1] / x[0])


def nonzero_dynamic_range(x):
    """Returns the ratio of largest and smallest non-zero values (by magnitude) in x (dB)

    Args:
        x (jax.numpy.ndarray): A sparse/compressible signal

    Returns:
        (float): The dynamic range between largest and smallest nonzero value

    See Also:
        :func:`dynamic_range`
    """
    x = jnp.sort(jnp.abs(x))
    idx = jnp.argmax(x != 0)
    return 20 * jnp.log10(x[-1] / x[idx])




# Copyright 2022 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Basic data compression routines
"""
# pylint: disable=W0611


from cr.nimble._src.compression.binary_arrs import (
    count_binary_runs,
    encode_binary_arr,
    decode_binary_arr,
    binary_compression_ratio,
    binary_space_saving_ratio,
)

from cr.nimble._src.compression.fixed_length import (
    encode_uint_arr_fl,
    decode_uint_arr_fl,
    encode_int_arr_sgn_mag_fl,
    decode_int_arr_sgn_mag_fl
)

from cr.nimble._src.compression.run_length import (
    count_runs_values,
    expand_runs_values
)

from cr.nimble._src.compression.bits import (
    float_to_int,
    int_to_float,
    int_to_bitarray,
    read_int_from_bitarray,
    float_to_bitarray,
    bitarray_to_float,
    read_float_from_bitarray
)



# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jax import jit
from jax.scipy import signal


def vec_convolve(x, h):
    """1D full convolution based on a hack suggested by Jake Vanderplas

    See https://github.com/google/jax/discussions/7961 for details
    """
    return signal.convolve(x[None], h[None])[0]

vec_convolve_jit = jit(vec_convolve)




# Copyright 2022 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Run Length Encoding


References:

https://stackoverflow.com/questions/1066758/find-length-of-sequences-of-identical-values-in-a-numpy-array-run-length-encodi
"""


import numpy as np

def count_runs_values(input_arr):
    """Computes run lengths of an array of integers
    """
    # make sure that input is a numpy array
    input_arr = np.asarray(input_arr)
    n = len(input_arr)
    if n == 0:
        return (np.empty(0),np.empty(0))
    if n == 1:
        return (np.ones(1), input_arr)
    # locate the changes
    changes = input_arr[1:] != input_arr[:-1]
    changes, = np.where(changes)
    # the last position should always be recorded as a change
    changes = np.append(changes, n-1)
    values = input_arr[changes]
    changes = np.insert(changes, 0, -1)
    # run lengths can be computed now
    runs = np.diff(changes)
    return runs, values

def expand_runs_values(runs, values):
    """Decodes run lengths to form an array of integers
    """
    return np.concatenate(
        [v * np.ones(r, dtype=np.int32) 
        for r, v in zip(runs, values)])



# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp
from jax import jit


def scale_0_mean_1_var(data, axis=-1):
    """Normalizes a data vector (data - mu) / sigma 

    Args:
        data (jax.numpy.ndarray): A data vector or array
        axis (int): For nd arrays, the axis along which the data normalization will be done

    Returns:
        (jax.numpy.ndarray, jax.numpy.ndarray, jax.numpy.ndarray): A tuple comprising of:
            * Normalized data vector/array
            * Mean value(s)
            * Standard deviation value(s)
    """
    mu = jnp.mean(data, axis)
    data = data - mu
    variance = jnp.var(data, axis)
    sigma = jnp.sqrt(variance)
    data = data / sigma
    return data, mu, sigma

scale_0_mean_1_var_jit = jit(scale_0_mean_1_var, static_argnums=(1,))


def scale_to_0_1(x):
    """Scales a signal to the range of 0 and 1

    Args:
        x (jax.numpy.ndarray): A signal to be scaled

    Returns:
        (jax.numpy.ndarray, float, float): A tuple comprising of:
            * Scaled signal
            * The amount of shift
            * The scale factor
    """
    shift = jnp.min(x)
    x = x - shift
    scale = jnp.max(x)
    x = x / scale
    return x, shift, scale

def descale_from_0_1(x, shift, scale):
    """Reverses the scaling of a signal from the range of 0 and 1

    Args:
        x (jax.numpy.ndarray): A signal to be scaled
        shift (float): The amount of shift
        scale (float): The scale factor

    Returns:
        jax.numpy.ndarray: Descaled signal
    """
    x = x * scale
    x = x + shift
    return x



# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
from jax import jit

def hard_threshold(x, K):
    """Returns the indices and corresponding values of largest K non-zero entries in a vector x

    Args:
        x (jax.numpy.ndarray): A sparse/compressible signal
        K (int): The number of largest entries to be kept in x

    Returns:
        (jax.numpy.ndarray, jax.numpy.ndarray): A tuple comprising of:
            * The indices of K largest entries in x
            * Corresponding entries in x

    See Also:
        :func:`hard_threshold_sorted`
        :func:`hard_threshold_by`
    """
    indices = jnp.argsort(jnp.abs(x))
    I = indices[:-K-1:-1]
    x_I = x[I]
    return I, x_I

def hard_threshold_sorted(x, K):
    """Returns the sorted indices and corresponding values of largest K non-zero entries in a vector x

    Args:
        x (jax.numpy.ndarray): A sparse/compressible signal
        K (int): The number of largest entries to be kept in x

    Returns:
        (jax.numpy.ndarray, jax.numpy.ndarray): A tuple comprising of:
            * The indices of K largest entries in x sorted in ascending order
            * Corresponding entries in x

    See Also:
        :func:`hard_threshold`
    """
    # Sort entries in x by their magnitude
    indices = jnp.argsort(jnp.abs(x))
    # Pick the indices of K-largest (magnitude) entries in x (from behind)
    I = indices[:-K-1:-1]
    # Make sure that indices are sorted in ascending order
    I = jnp.sort(I)
    # Pick corresponding values
    x_I = x[I]
    return I, x_I

def hard_threshold_by(x, t):
    """
    Sets all entries in x to be zero which are less than t in magnitude

    Args:
        x (jax.numpy.ndarray): A sparse/compressible signal
        t (float): The threshold value

    Returns:
        (jax.numpy.ndarray): x modified such that all values below t are set to 0

    Note:
        This function doesn't change the length of x and can be JIT compiled

    See Also:
        :func:`hard_threshold`
    """
    valid = jnp.abs(x) >= t
    return x * valid

def largest_indices_by(x, t):
    """
    Returns the locations of all entries in x which are larger than t in magnitude

    Args:
        x (jax.numpy.ndarray): A sparse/compressible signal
        t (float): The threshold value

    Returns:
        (jax.numpy.ndarray): An index vector of all entries in x which are above the threshold

    Note:
        This function cannot be JIT compiled as the length of output is data dependent

    See Also:
        :func:`hard_threshold_by`
    """
    return jnp.where(jnp.abs(x) >= t)[0]


def energy_threshold(signal, fraction):
    """
    Keeps only as much coefficients in signal so as to capture a fraction of signal energy 

    Args:
        x (jax.numpy.ndarray): A signal
        fraction (float): The fraction of energy to be preserved

    Returns:
        (jax.numpy.ndarray, jax.numpy.ndarray): A tuple comprising of:
            * Signal after thresholding
            * A binary mask of the indices to be kept

    Note:
        This function doesn't change the length of signal and can be JIT compiled

    See Also:
        :func:`hard_threshold`
    """
    # signal length
    n = signal.size
    # compute energies
    energies = signal ** 2
    # sort in descending order
    idx = jnp.argsort(energies)[::-1]
    energies = energies[idx]
    # total energy
    s = jnp.sum(energies) * 1.
    # normalize
    energies = energies / s
    # convert to a cmf
    cmf = jnp.cumsum(energies)
    # find the index
    index =  jnp.argmax(cmf >= fraction)
    # build the mask
    idx2 = jnp.arange(n)
    mask = jnp.where(idx2 <= index, 1, 0)
    # reshuffle the mask
    mask = mask.at[idx].set(mask)
    signal = signal * mask
    return signal, mask



# Copyright 2022-Present CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Synthetic Signals
"""
# pylint: disable=W0611

from cr.nimble._src.dsp.synthetic_signals import (
    chirp,
    chirp_centered,
    pulse,
    gaussian,
    gaussian_pulse,
    decaying_sine_wave,
    transient_sine_wave
)


# Copyright 2022 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Fixed Length Encoding of arrays
"""

from bitarray import bitarray
from bitarray.util import int2ba, ba2int
import numpy as np


def encode_uint_arr_fl(input_arr, bits_per_sample: int):
    """Encodes an array of unsigned integers to a bit array using a fixed number of bits per sample
    """
    a = bitarray()
    max_val = (1 << bits_per_sample) - 1
    # make sure that the values are clipped
    input_arr = np.clip(input_arr, 0, max_val)
    for value in input_arr:
        value = int(value)
        a.extend(int2ba(value, bits_per_sample))
    return a

def decode_uint_arr_fl(input_bit_arr : bitarray, bits_per_sample: int):
    """Decodes an array of unsigned integers from a bit array using a fixed number of bits per sample
    """
    a = input_bit_arr
    # number of bits
    nbits = len(a)
    # number of samples
    n = nbits // bits_per_sample
    output = np.empty(n, dtype=np.int)
    idx = 0
    for i in range(n):
        # read the value
        value = ba2int(a[idx:idx+bits_per_sample])
        idx += bits_per_sample
        output[i] = value
    return output


def encode_int_arr_sgn_mag_fl(input_arr, bits_per_sample: int):
    """Encodes an array of integers to a bit array using a sign bit and a fixed number of bits per sample for magnitude
    """
    a = bitarray()
    max_val = (1 << bits_per_sample) - 1
    for value in input_arr:
        value = int(value)
        sign, value = (0, value) if value >= 0 else (1, -value)
        # make sure that the values are clipped
        value = max_val if value > max_val else value
        a.append(sign)
        a.extend(int2ba(value, bits_per_sample))
    return a


def decode_int_arr_sgn_mag_fl(input_bit_arr : bitarray, bits_per_sample: int):
    """Decodes an array of integers from a bit array using a sign bit and a fixed number of bits per sample for magnitude
    """
    a = input_bit_arr
    # number of bits
    nbits = len(a)
    # number of samples
    n = nbits // (bits_per_sample + 1)
    output = np.empty(n, dtype=np.int)
    idx = 0
    for i in range(n):
        # read the sign bit
        s = a[idx]
        idx += 1
        # read the value
        value = ba2int(a[idx:idx+bits_per_sample])
        idx += bits_per_sample
        # combine sign and value
        value = -value if s else value
        output[i] = value
    return output




# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
from jax import jit


def quantize_1(x, n):
    """Quantizes a signal to n bits where signal values are bounded by 1

    Args:
        x (jax.numpy.ndarray): A signal to be quantized
        n (int): number of bits for quantization

    Returns:
        (jax.numpy.ndarray): Quantized signal with integer values
    """
    # scaling
    factor = 2**n-1
    x = factor * x
    # quantization
    x = jnp.round(x)
    # type conversion from float to int
    x = x.astype(int)
    return x


def inv_quantize_1(x, n):
    """Inverse quantizes a signal from n bits

    Args:
        x (jax.numpy.ndarray): A signal to be inverse quantized
        n (int): number of bits for quantization

    Returns:
        (jax.numpy.ndarray): Quantized signal with integer values
    """
    # scaling
    factor = 2**n-1
    x = x / factor
    return x



# Copyright 2021 CR.Sparse Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
from jax import jit

from cr.nimble import (
    is_matrix,
    sqr_norms_l2_rw,
    sqr_norms_l2_cw)

def find_first_signal_with_energy_le_rw(X, energy):
    """Returns the index of the first row which has energy less than the specified threshold
    """
    assert is_matrix(X)
    energies = sqr_norms_l2_rw(X)
    index = jnp.argmax(energies <= energy)
    return index if energies[index] <= energy else jnp.array(-1)

def find_first_signal_with_energy_le_cw(X, energy):
    """Returns the index of the first column which has energy less than the specified threshold
    """
    assert is_matrix(X)
    energies = sqr_norms_l2_cw(X)
    index = jnp.argmax(energies <= energy)
    return index if energies[index] <= energy else jnp.array(-1)


def energy(data, axis=-1):
    """
    Computes the energy of the signal along the specified axis
    """
    power = jnp.abs(data) ** 2
    return jnp.sum(power, axis)





# Copyright 2022 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import struct
from bitarray import bitarray
from bitarray.util import int2ba, ba2int


def float_to_int(value):
    """Constructs an integer representation of a floating point number
    """
    s = struct.pack('>f', value)
    return struct.unpack('>l', s)[0]


def int_to_float(rep):
    """Constructs a floating point value from an integer representation
    """
    s = struct.pack('>l', rep)
    return struct.unpack('>f', s)[0]


def int_to_bitarray(value, len_bits=5):
    n = value.bit_length()
    ba = int2ba(value, length=n+1, signed=True)
    # some-bits to encode the length of integer value
    output = int2ba(n+1, length=len_bits)
    # now add the integer bit array
    output.extend(ba)
    return output

def read_int_from_bitarray(a: bitarray, pos:int, len_bits:int =5):
    # read the six bit prefix
    e = pos + len_bits
    prefix = a[pos:e]
    n = ba2int(prefix)
    pos = e
    e = e + n
    suffix = a[pos:e]
    value = ba2int(suffix, signed=True)
    return value, e


def float_to_bitarray(value):
    s = struct.pack('>f', value)
    ba = bitarray()
    ba.frombytes(s)
    return ba

def bitarray_to_float(a: bitarray):
    bytes = a.tobytes()
    value = struct.unpack('>f', bytes)[0]
    return value    

def read_float_from_bitarray(a: bitarray, pos: int):
    e = pos+32
    value = a[pos:e]
    return bitarray_to_float(value), e

